/* This solution traverses rows (for red movement) and columns (for
   blue movement) in such a way that it can overwrite the array as
   it's processed. The last element of a row/column must be handled
   specially, since the first row/column might be vacated in the same
   step. 

   The use of OMP is a little simpler and more direct than in the
   alternate solution. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bsd_rand.h"

#define N 100
#define T 10
#define C 0.5

char m[N][N];

static void rand_fill(char c, int n)
{
  while (n--) {
    int i, j;
    do {
      i = bsd_rand(N);
      j = bsd_rand(N);
    } while (m[i][j] != ' ');
    m[i][j] = c;
  }
}

static void init_array()
{
  memset(m, ' ', sizeof(m));
  rand_fill('r', N * N / 4);
  rand_fill('b', N * N / 4);
}

static void print_array()
{
  int i, j;
  char line[N + 1];

  for (j = 0; j < N; j++) {
    for (i = 0; i < N; i++) {
      line[i] = m[i][j];
    }
    line[N] = '\n';
    fwrite(line, N+1, 1, stdout);
  }
}

static int count_cells(int c, int ti, int tj)
{
  int i, j, cnt = 0;

  ti *= T;
  tj *= T;

  for (i = 0; i < T; i++)
    for (j = 0; j < T; j++) {
      if (m[ti + i][tj + j] == c)
        cnt++;
    }

  return cnt;
}

static int check_regions()
{
  int i, done = 0;

  #pragma omp parallel for reduction(||:done)
  for (i = 0; i < N / T; i++) {
    int j;
    for (j = 0; !done && j < N / T; j++) {
      int rc = count_cells('r', i, j);
      int bc = count_cells('b', i, j);

      if ((rc > (T * T * C))
          || (bc > (T * T * C)))
        done = 1;
    }
  }

  return done;
}

int main(int argc, char **argv)
{
  int n, cnt = 0, show = 0, red_moved, blue_moved;

  if (argc < 2) {
    fprintf(stderr, "need an argument\n");
    return 1;
  }

  show = argc > 2;
    
  n = atoi(argv[1]);
  bsd_srand(n);
  
  init_array();

  if (show)
    printf("%d\n", N);

  do {
    int i, j;

    if (show)
      print_array();

    if (check_regions())
      break;

    red_moved = 0;

    #pragma omp parallel for default(none) shared(m) private(i) reduction(||:red_moved)
    for (j = 0; j < N; j++) {
      int left_white = (m[0][j] == ' ');
      for (i = 0; i < N-1; i++) {
        if ((m[i][j] == 'r') && (m[i+1][j] == ' ')) {
          m[i][j] = ' ';
          m[i+1][j] = 'r';
          red_moved = 1;
          i++; /* don't move again */
        }
      }
      if (i < N) {
        if ((m[N-1][j] == 'r') && left_white) {
          m[N-1][j] = ' ';
          m[0][j] = 'r';
          red_moved = 1;
        }
      }
    }

    blue_moved = 0;

    #pragma omp parallel for default(none) shared(m) private(j) reduction(||:blue_moved)
    for (i = 0; i < N; i++) {
      int top_white = (m[i][0] == ' ');
      for (j = 0; j < N-1; j++) {
        if ((m[i][j] == 'b') && (m[i][j+1] == ' ')) {
          m[i][j] = ' ';
          m[i][j+1] = 'b';
          blue_moved = 1;
          j++; /* don't move again */
        }
      }
      if (j < N) {
        if ((m[i][N-1] == 'b') && top_white) {
          m[i][N-1] = ' ';
          m[i][0] = 'b';
          blue_moved = 1;
        }
      }
    }

    cnt++;
  } while (red_moved || blue_moved);

  if (!show)
    printf("%d\n", cnt);

  return 0;
}
