/* Simpler solution. Keeps the old and new arrays for each
   half of the step, so there's no issue of overwriting data
   that is needed later.

   The use of OMP is a little less simple than the other solution.  It
   illustrate how `single' and `barrier' might be used. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bsd_rand.h"

#define N 512
#define T 32
#define C 0.5

char m[N][N], tmp[N][N];

static void rand_fill(char c, int n)
{
  while (n--) {
    int i, j;
    do {
      i = bsd_rand(N);
      j = bsd_rand(N);
    } while (m[i][j] != ' ');
    m[i][j] = c;
  }
}

static void init_array()
{
  memset(m, ' ', sizeof(m));
  rand_fill('r', N * N / 4);
  rand_fill('b', N * N / 4);
}

static void print_array()
{
  int i, j;
  char line[N + 1];

  for (j = 0; j < N; j++) {
    for (i = 0; i < N; i++) {
      line[i] = m[i][j];
    }
    line[N] = '\n';
    fwrite(line, N+1, 1, stdout);
  }
}

static int count_cells(int c, int ti, int tj)
{
  int i, j, cnt = 0;

  ti *= T;
  tj *= T;

  for (i = 0; i < T; i++)
    for (j = 0; j < T; j++) {
      if (m[ti + i][tj + j] == c)
        cnt++;
    }

  return cnt;
}

static int check_regions()
{
  int i, done = 0;

  #pragma omp parallel for reduction(||:done)
  for (i = 0; i < N / T; i++) {
    int j;
    for (j = 0; !done && j < N / T; j++) {
      int rc = count_cells('r', i, j);
      int bc = count_cells('b', i, j);

      if ((rc > (T * T * C))
          || (bc > (T * T * C)))
        done = 1;
    }
  }

  return done;
}

int main(int argc, char **argv)
{
  int n, cnt = 0, show = 0, moved;

  if (argc < 2) {
    fprintf(stderr, "need an argument\n");
    return 1;
  }

  show = argc > 2;
    
  n = atoi(argv[1]);
  bsd_srand(n);
  
  init_array();

  if (show)
    printf("%d\n", N);

  do {
    int i, j;

    if (show)
      print_array();

    if (check_regions())
      break;

    memset(tmp, ' ', sizeof(tmp));

    moved = 0;

    #pragma omp parallel default(none) private(j) shared(m, tmp) reduction(||:moved)
    {
      #pragma omp for
      for (i = 0; i < N; i++) {
        for (j = 0; j < N; j++) {
          switch (m[i][j]) {
          case 'r':
            {
              int i2 = (i + 1) % N;
              if (m[i2][j] == ' ') {
                tmp[i2][j] = 'r';
                moved = 1;
              } else {
                tmp[i][j] = 'r';
              }
            }
            break;
          case 'b':
            tmp[i][j] = 'b';
            break;
          }
        }
      }

      #pragma omp barrier
      #pragma omp single
      memset(m, ' ', sizeof(m));
      #pragma omp barrier

      #pragma omp for
      for (i = 0; i < N; i++) {
        for (j = 0; j < N; j++) {
          switch (tmp[i][j]) {
          case 'r':
            m[i][j] = 'r';
            break;
          case 'b':
            {
              int j2 = (j + 1) % N;
              if (tmp[i][j2] == ' ') {
                m[i][j2] = 'b';
                moved = 1;
              } else {
                m[i][j] = 'b';
              }
            }
            break;
          }
        }
      }
    }

    cnt++;
  } while (moved);

  if (!show)
    printf("%d\n", cnt);

  return 0;
}
